package sound.spectrogram;

import sound.soundDemo.JavaSound;
import sound.soundDemo.OpenCloseInterface;
import sound.OscopePanel;
import sound.SoundUtils;
import sound.CapturePlayBackPanel;

import javax.sound.sampled.*;
import javax.swing.*;
import javax.swing.border.BevelBorder;
import javax.swing.border.CompoundBorder;
import javax.swing.border.EmptyBorder;
import java.awt.*;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.awt.font.FontRenderContext;
import java.awt.font.LineBreakMeasurer;
import java.awt.font.TextAttribute;
import java.awt.font.TextLayout;
import java.awt.geom.Line2D;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.text.AttributedCharacterIterator;
import java.text.AttributedString;
import java.util.Enumeration;
import java.util.Vector;

import gui.run.RunButton;

/**
 * Modified from the Brian Lichtenwalter version  by R distinti
 */
public class SimplifiedCapturePlayBackPanel extends JPanel
        implements ActionListener, OpenCloseInterface {
    private final int bufSize = 16384;
    private Capture capture = new Capture();
    private Playback playback = new Playback();
    private int[] audioData = {0};
    //private UlawCodec ulc = new UlawCodec();
    OscopePanel osp = new OscopePanel();
    private AudioInputStream audioInputStream;
    private JButton playB, captB, pausB, doneB;
    private JTextField textField;
    private String fileName = "untitled";
    private String errStr;
    private double duration, seconds;
    private File file;
    private Vector lines = new Vector();
    private JDialog myParent;

    public SimplifiedCapturePlayBackPanel(JDialog parent) {
        setLayout(new BorderLayout());
        add(osp, BorderLayout.CENTER);
        add(getButtons(), BorderLayout.SOUTH);
        myParent=parent;
    }

    public JPanel getButtons() {
        JPanel buttonsPanel = new JPanel();
        buttonsPanel.setLayout(new FlowLayout());
        playB = addButton("Play", buttonsPanel, false);
        captB = addButton("Record", buttonsPanel, true);
        pausB = addButton("Pause", buttonsPanel, false);
        buttonsPanel.add(doneB=new RunButton("Done"){
            public void run(){
                 myParent.setVisible(false);
            }
        });
        return buttonsPanel;
    }

    public void open() {
    }

    public void close() {
        if (playback.thread != null) {
            playB.doClick(0);
        }
        if (capture.thread != null) {
            captB.doClick(0);
        }
    }

    private JButton addButton(String name, JPanel p, boolean state) {
        JButton b = new JButton(name);
        b.addActionListener(this);
        b.setEnabled(state);
        p.add(b);
        return b;
    }

    public void actionPerformed(ActionEvent e) {
        Object obj = e.getSource();
        if (obj.equals(playB)) {
            if (playB.getText().startsWith("Play")) {
                playback.start();
                // samplingGraph.start();
                captB.setEnabled(false);
                pausB.setEnabled(true);
                playB.setText("Stop");
            } else {
                playback.stop();
                captB.setEnabled(true);
                pausB.setEnabled(false);
                playB.setText("Play");
            }
        } else if (obj.equals(captB)) {
            if (captB.getText().startsWith("Record")) {
                file = null;
                capture.start();
                fileName = "untitled";
                // samplingGraph.start();
                playB.setEnabled(false);
                pausB.setEnabled(true);
                doneB.setEnabled(false);
                captB.setText("Stop");
            } else {
                lines.removeAllElements();
                capture.stop();
                playB.setEnabled(true);
                pausB.setEnabled(false);
                doneB.setEnabled(true);
                captB.setText("Record");
                getUpdateDisplay();
            }
        } else if (obj.equals(pausB)) {
            if (pausB.getText().startsWith("Pause")) {
                if (capture.thread != null) {
                    capture.line.stop();
                } else {
                    if (playback.thread != null) {
                        playback.line.stop();
                    }
                }
                pausB.setText("Resume");
            } else {
                if (capture.thread != null) {
                    capture.line.start();
                } else {
                    if (playback.thread != null) {
                        playback.line.start();
                    }
                }
                pausB.setText("Pause");
            }
        }
    }

    public int[] getData() {
        return audioData;
    }

    public void createWaveForm(byte[] audioBytes) {
        AudioFormat format = audioInputStream.getFormat();
        System.out.println("format=" + format);
        if (audioBytes == null) {
            try {
                audioBytes = new byte[
                        (int) (audioInputStream.getFrameLength()
                        * format.getFrameSize())];
                audioInputStream.read(audioBytes);
            } catch (Exception ex) {
                reportStatus(ex.toString());
                return;
            }
        }
        Dimension d = getSize();
        if (format.getSampleSizeInBits() == 16) {
            int nlengthInSamples = audioBytes.length / 2;
            audioData = new int[nlengthInSamples];
            System.out.println(audioData.length);
            System.out.println("audio 1 " + audioData[10]);
            if (format.isBigEndian()) {
                for (int i = 0; i < nlengthInSamples; i++) {
                    /* First byte is MSB (high order) */
                    int MSB = (int) audioBytes[2 * i];
                    /* Second byte is LSB (low order) */
                    int LSB = (int) audioBytes[2 * i + 1];
                    audioData[i] = MSB << 8 | (255 & LSB);
                }
            } else {
                for (int i = 0; i < nlengthInSamples; i++) {
                    /* First byte is LSB (low order) */
                    int LSB = (int) audioBytes[2 * i];
                    /* Second byte is MSB (high order) */
                    int MSB = (int) audioBytes[2 * i + 1];
                    audioData[i] = MSB << 8 | (255 & LSB);
                }
            }
        } else if (format.getSampleSizeInBits() == 8) {
            int nlengthInSamples = audioBytes.length;
            audioData = new int[nlengthInSamples];
            if (format.getEncoding().toString().startsWith("PCM_SIGN")) {
                for (int i = 0; i < audioBytes.length; i++) {
                    audioData[i] = audioBytes[i];
                }
            } else {
                for (int i = 0; i < audioBytes.length; i++) {
                    audioData[i] = audioBytes[i] - 128;
                }
            }
        }
        osp.setData(getDoubleData(audioData));
    }

    /**
     * Convert the 16 bit data into double data
     *
     * @param audioData
     * @return data that ranges from -1 to 1
     */
    private double[] getDoubleData(int[] audioData) {
        double d[] = new double[audioData.length];
        for (int i = 0; i < audioData.length; i++)
            d[i] = audioData[i] / 32768.0;
        return d;
    }

    private void getUpdateDisplay() {
    }

    public void createAudioInputStream(File file,
                                       boolean updateComponents) {
        if (file != null && file.isFile()) {
            try {
                this.file = file;
                errStr = null;
                audioInputStream = AudioSystem.getAudioInputStream(file);
                playB.setEnabled(true);
                fileName = file.getName();
                long milliseconds = (long) ((audioInputStream.getFrameLength() *
                        1000) /
                        audioInputStream.getFormat().getFrameRate());
                duration = milliseconds / 1000.0;
            } catch (Exception ex) {
                reportStatus(ex.toString());
            }
        } else {
            reportStatus("Audio file required.");
        }
    }

    public void saveToFile(String name, AudioFileFormat.Type fileType) {
        if (audioInputStream == null) {
            reportStatus("No loaded audio to save");
            return;
        } else if (file != null) {
            createAudioInputStream(file, false);
        }

        // reset to the beginnning of the captured data
        try {
            audioInputStream.reset();
        } catch (Exception e) {
            reportStatus("Unable to reset stream " + e);
            return;
        }
        File file = new File(fileName = name);
        try {
            if (AudioSystem.write(audioInputStream, fileType, file) == -1) {
                throw new IOException("Problems writing to file");
            }
        } catch (Exception ex) {
            reportStatus(ex.toString());
        }
    }

    private void reportStatus(String msg) {
        if ((errStr = msg) != null) {
            System.out.println(errStr);
        }
    }

    /**
     * Write data to the OutputChannel.
     */
    public class Playback implements Runnable {
        SourceDataLine line;
        Thread thread;

        public void start() {
            errStr = null;
            thread = new Thread(this);
            thread.setName("Playback");
            thread.start();
        }

        public void stop() {
            thread = null;
        }

        private void shutDown(String message) {
            if ((errStr = message) != null) {
                System.err.println(errStr);
            }
            if (thread != null) {
                thread = null;
                captB.setEnabled(true);
                pausB.setEnabled(false);
                playB.setText("Play");
            }
        }

        public void run() {
            // reload the file if loaded by file
            if (file != null) {
                createAudioInputStream(file, false);
            }

            // make sure we have something to play
            if (audioInputStream == null) {
                shutDown("No loaded audio to play back");
                return;
            }
            // reset to the beginnning of the stream
            try {
                audioInputStream.reset();
            } catch (Exception e) {
                shutDown("Unable to reset the stream\n" + e);
                return;
            }

            // get an AudioInputStream of the desired format for playback
            AudioFormat format = getFormat();
            AudioInputStream playbackInputStream = AudioSystem.getAudioInputStream(
                    format, audioInputStream);
            if (playbackInputStream == null) {
                shutDown(
                        "Unable to convert stream of format " +
                        audioInputStream +
                        " to format " +
                        format);
                return;
            }

            // define the required attributes for our line,
            // and make sure a compatible line is supported.

            DataLine.Info info = new DataLine.Info(SourceDataLine.class,
                    format);
            if (!AudioSystem.isLineSupported(info)) {
                shutDown("Line matching " + info + " not supported.");
                return;
            }

            // get and open the source data line for playback.

            try {
                line = (SourceDataLine) AudioSystem.getLine(info);
                line.open(format, bufSize);
            } catch (LineUnavailableException ex) {
                shutDown("Unable to open the line: " + ex);
                return;
            }

            // play back the captured audio data

            int frameSizeInBytes = format.getFrameSize();
            int bufferLengthInFrames = line.getBufferSize() / 8;
            int bufferLengthInBytes = bufferLengthInFrames *
                    frameSizeInBytes;
            byte[] data = new byte[bufferLengthInBytes];
            int numBytesRead = 0;

            // start the source data line
            line.start();
            while (thread != null) {
                try {
                    if ((numBytesRead = playbackInputStream.read(data)) ==
                            -1) {
                        break;
                    }
                    int numBytesRemaining = numBytesRead;
                    while (numBytesRemaining > 0) {
                        numBytesRemaining -=
                                line.write(data, 0, numBytesRemaining);
                    }
                } catch (Exception e) {
                    shutDown("Error during playback: " + e);
                    break;
                }
            }
            // we reached the end of the stream.  let the data play out, then
            // stop and close the line.
            if (thread != null) {
                line.drain();
            }
            line.stop();
            line.close();
            line = null;
            shutDown(null);
        }
    } // End class Playback

    /**
     * Reads data from the input channel and writes to the output stream
     */
    class Capture implements Runnable {
        TargetDataLine line;
        Thread thread;

        public void start() {
            errStr = null;
            thread = new Thread(this);
            thread.setName("Capture");
            thread.start();
        }

        public void stop() {
            thread = null;
        }

        private void shutDown(String message) {
            if ((errStr = message) != null && thread != null) {
                thread = null;
                playB.setEnabled(true);
                pausB.setEnabled(false);
                captB.setText("Record");
                System.err.println(errStr);
            }
        }

        public void run() {
            duration = 0;
            audioInputStream = null;

            // define the required attributes for our line,
            // and make sure a compatible line is supported.

            AudioFormat format = getFormat();
            DataLine.Info info = new DataLine.Info(TargetDataLine.class,
                    format);
            if (!AudioSystem.isLineSupported(info)) {
                shutDown("Line matching " + info + " not supported.");
                return;
            }

            // get and open the target data line for capture.

            try {
                line = (TargetDataLine) AudioSystem.getLine(info);
                line.open(format, line.getBufferSize());
            } catch (LineUnavailableException ex) {
                shutDown("Unable to open the line: " + ex);
                return;
            } catch (SecurityException ex) {
                shutDown(ex.toString());
                SoundUtils.showInfoDialog();
                return;
            } catch (Exception ex) {
                shutDown(ex.toString());
                return;
            }

            // play back the captured audio data
            ByteArrayOutputStream out = new ByteArrayOutputStream();
            int frameSizeInBytes = format.getFrameSize();
            int bufferLengthInFrames = line.getBufferSize() / 8;
            int bufferLengthInBytes = bufferLengthInFrames *
                    frameSizeInBytes;
            byte[] data = new byte[bufferLengthInBytes];
            int numBytesRead;
            line.start();
            while (thread != null) {
                if ((numBytesRead =
                        line.read(data, 0, bufferLengthInBytes)) ==
                        -1) {
                    break;
                }
                out.write(data, 0, numBytesRead);
            }

            // we reached the end of the stream.  stop and close the line.
            line.stop();
            line.close();
            line = null;

            // stop and close the output stream
            try {
                out.flush();
                out.close();
            } catch (IOException ex) {
                ex.printStackTrace();
            }

            // load bytes into the audio input stream for playback

            byte audioBytes[] = out.toByteArray();
            ByteArrayInputStream bais = new ByteArrayInputStream(
                    audioBytes);
            audioInputStream =
                    new AudioInputStream(bais,
                            format,
                            audioBytes.length / frameSizeInBytes);
            long milliseconds = (long) ((audioInputStream.getFrameLength() *
                    1000) /
                    format.getFrameRate());
            duration = milliseconds / 1000.0;
            try {
                audioInputStream.reset();
            } catch (Exception ex) {
                ex.printStackTrace();
                return;
            }
            createWaveForm(audioBytes);
        }
    } // End class Capture
    public AudioFormat getFormat() {
         return new AudioFormat(8000,
                16,
                1,
                true,
                true);
    }

    /**
     * Render a WaveForm.
     */
    class SamplingGraph extends JPanel implements Runnable {
        private Thread thread;
        //private Font font10 = new Font("serif", Font.PLAIN, 10);
        private Font font12 = new Font("serif", Font.PLAIN, 12);
        Color jfcBlue = new Color(204, 204, 255);
        Color pink = new Color(255, 175, 175);

        public SamplingGraph() {
            setBackground(new Color(20, 20, 20));
        }

        public void createWaveForm(byte[] ulawEncodedData) {
            System.out.println("WaveForm");
            lines.removeAllElements();  // clear the old vector
            AudioFormat format = audioInputStream.getFormat();
            if (ulawEncodedData == null) {
                try {
                    ulawEncodedData = new byte[
                            (int) (audioInputStream.getFrameLength()
                            * format.getFrameSize())];
                    audioInputStream.read(ulawEncodedData);
                } catch (Exception ex) {
                    reportStatus(ex.toString());
                    return;
                }
            }
            Dimension d = getSize();
            int w = d.width;
            int h = d.height - 15;
            if (format.getSampleSizeInBits() == 16) {
                int nlengthInSamples = ulawEncodedData.length / 2;
                audioData = new int[nlengthInSamples];
                if (format.isBigEndian()) {
                    for (int i = 0; i < nlengthInSamples; i++) {
                        /* First byte is MSB (high order) */
                        int MSB = (int) ulawEncodedData[2 * i];
                        /* Second byte is LSB (low order) */
                        int LSB = (int) ulawEncodedData[2 * i + 1];
                        audioData[i] = MSB << 8 | (255 & LSB);
                    }
                } else {
                    for (int i = 0; i < nlengthInSamples; i++) {
                        /* First byte is LSB (low order) */
                        int LSB = (int) ulawEncodedData[2 * i];
                        /* Second byte is MSB (high order) */
                        int MSB = (int) ulawEncodedData[2 * i + 1];
                        audioData[i] = MSB << 8 | (255 & LSB);
                    }
                }
            } else if (format.getSampleSizeInBits() == 8) {
                int nlengthInSamples = ulawEncodedData.length;
                audioData = new int[nlengthInSamples];
                if (format.getEncoding().toString().startsWith("PCM_SIGN")) {
                    for (int i = 0; i < ulawEncodedData.length; i++) {
                        audioData[i] = ulawEncodedData[i];
                    }
                } else {
                    for (int i = 0; i < ulawEncodedData.length; i++) {
                        audioData[i] = ulawEncodedData[i] - 128;
                    }
                }
            }
            int frames_per_pixel = ulawEncodedData.length /
                    format.getFrameSize() /
                    w;
            byte my_byte = 0;
            double y_last = 0;
            int numChannels = format.getChannels();
            for (double x = 0; x < w && audioData != null; x++) {
                int idx = (int) (frames_per_pixel * numChannels * x);
                if (format.getSampleSizeInBits() == 8) {
                    my_byte = (byte) audioData[idx];
                } else {
                    my_byte = (byte) (128 * audioData[idx] / 32768);
                }
                double y_new = (double) (h * (128 - my_byte) / 256);
                lines.add(new Line2D.Double(x, y_last, x, y_new));
                y_last = y_new;
            }
            repaint();
        }

        public void paint(Graphics g) {
            Dimension d = getSize();
            int w = d.width;
            int h = d.height;
            int INFOPAD = 15;
            Graphics2D g2 = (Graphics2D) g;
            g2.setBackground(getBackground());
            g2.clearRect(0, 0, w, h);
            g2.setColor(Color.white);
            g2.fillRect(0, h - INFOPAD, w, INFOPAD);
            if (errStr != null) {
                g2.setColor(jfcBlue);
                g2.setFont(new Font("serif", Font.BOLD, 18));
                g2.drawString("ERROR", 5, 20);
                AttributedString as = new AttributedString(errStr);
                as.addAttribute(TextAttribute.FONT,
                        font12,
                        0,
                        errStr.length());
                AttributedCharacterIterator aci = as.getIterator();
                FontRenderContext frc = g2.getFontRenderContext();
                LineBreakMeasurer lbm = new LineBreakMeasurer(aci, frc);
                float x = 5, y = 25;
                lbm.setPosition(0);
                while (lbm.getPosition() < errStr.length()) {
                    TextLayout tl = lbm.nextLayout(w - x - 5);
                    if (!tl.isLeftToRight()) {
                        x = w - tl.getAdvance();
                    }
                    tl.draw(g2, x, y += tl.getAscent());
                    y += tl.getDescent() + tl.getLeading();
                }
            } else if (capture.thread != null) {
                g2.setColor(Color.black);
                g2.setFont(font12);
                g2.drawString("Length: " + String.valueOf(seconds),
                        3,
                        h - 4);
            } else {
                g2.setColor(Color.black);
                g2.setFont(font12);
                g2.drawString(
                        "File: " +
                        fileName +
                        "  Length: " +
                        String.valueOf(duration) +
                        "  Position: " +
                        String.valueOf(seconds),
                        3,
                        h - 4);
                if (audioInputStream != null) {
                    // .. render sampling graph ..
                    g2.setColor(jfcBlue);
                    for (int i = 1; i < lines.size(); i++) {
                        g2.draw((Line2D) lines.get(i));
                    }

                    // .. draw current position ..
                    if (seconds != 0) {
                        double loc = seconds / duration * w;
                        g2.setColor(pink);
                        g2.setStroke(new BasicStroke(3));
                        g2.draw(
                                new Line2D.Double(loc,
                                        0,
                                        loc,
                                        h - INFOPAD - 2));
                    }
                }
            }
        }

        public void start() {
            thread = new Thread(this);
            thread.setName("SamplingGraph");
            thread.start();
            seconds = 0;
        }

        public void stop() {
            if (thread != null) {
                thread.interrupt();
            }
            thread = null;
        }

        public void run() {
            seconds = 0;
            while (thread != null) {
                if ((playback.line != null) && (playback.line.isOpen())) {
                    long milliseconds = (playback.line.getMicrosecondPosition() /
                            1000);
                    seconds = milliseconds / 1000.0;
                } else if ((capture.line != null) &&
                        (capture.line.isActive())) {
                    long milliseconds = (capture.line.getMicrosecondPosition() /
                            1000);
                    seconds = milliseconds / 1000.0;
                }
                try {
                    Thread.sleep(100);
                } catch (Exception e) {
                    break;
                }
                repaint();
                while ((capture.line != null && !capture.line.isActive()) ||
                        (playback.line != null && !playback.line.isOpen())) {
                    try {
                        Thread.sleep(10);
                    } catch (Exception e) {
                        break;
                    }
                }
            }
            seconds = 0;
            repaint();
        }
    } // End class SamplingGraph


    public int getBufSize() {
        return bufSize;
    }


    public Capture getCapture() {
        return capture;
    }

    public Playback getPlayback() {
        return playback;
    }

    public AudioInputStream getAudioInputStream() {
        return audioInputStream;
    }

    public JButton getPlayB() {
        return playB;
    }

    public JButton getCaptB() {
        return captB;
    }

    public JButton getPausB() {
        return pausB;
    }

    public JTextField getTextField() {
        return textField;
    }

    public String getFileName() {
        return fileName;
    }

    public String getErrStr() {
        return errStr;
    }

    public double getDuration() {
        return duration;
    }

    public double getSeconds() {
        return seconds;
    }

    public File getFile() {
        return file;
    }

    public Vector getLines() {
        return lines;
    }
}